#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Interfaces describing the events and fields this package uses.

Also utility functions.
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import schema
from zope import interface

from zope.schema import interfaces as sch_interfaces

class IBeforeSchemaFieldAssignedEvent(interface.Interface):
    """
    An event sent when certain schema fields will be assigning
    an object to a property of another object.

    The interface :class:`.IBeforeObjectAssignedEvent` is a sub-interface
    of this one.
    """
    object = interface.Attribute("The object that is going to be assigned. Subscribers may modify this")

    name = interface.Attribute("The name of the attribute under which the object will be assigned.")

    context = interface.Attribute("The context object where the object will be assigned to.")

# Make this a base of the zope interface so our handlers
# are compatible
sch_interfaces.IBeforeObjectAssignedEvent.__bases__ = (IBeforeSchemaFieldAssignedEvent,)

@interface.implementer(IBeforeSchemaFieldAssignedEvent)
class BeforeSchemaFieldAssignedEvent(object):

    def __init__(self, obj, name, context):
        self.object = obj
        self.name = name
        self.context = context

class IBeforeTextAssignedEvent(IBeforeSchemaFieldAssignedEvent):
    """
    Event for assigning text.
    """

    object = schema.Text(title="The text being assigned.")

class IBeforeTextLineAssignedEvent(IBeforeTextAssignedEvent):  # ITextLine extends IText
    """
    Event for assigning text lines.
    """

    object = schema.TextLine(title="The text being assigned.")

class IBeforeContainerAssignedEvent(IBeforeSchemaFieldAssignedEvent):
    """
    Event for assigning containers (__contains__).
    """

class IBeforeIterableAssignedEvent(IBeforeContainerAssignedEvent):
    """
    Event for assigning iterables.
    """

class IBeforeCollectionAssignedEvent(IBeforeIterableAssignedEvent):
    """
    Event for assigning collections.
    """

    object = interface.Attribute("The collection being assigned. May or may not be mutable.")

class IBeforeSetAssignedEvent(IBeforeCollectionAssignedEvent):
    """
    Event for assigning sets.
    """

class IBeforeSequenceAssignedEvent(IBeforeCollectionAssignedEvent):
    """
    Event for assigning sequences.
    """

    object = interface.Attribute("The sequence being assigned. May or may not be mutable.")

class IBeforeDictAssignedEvent(IBeforeIterableAssignedEvent):
    """
    Event for assigning dicts.
    """

# The hierarchy is IContainer > IIterable > ICollection > ISequence > [ITuple, IList]
# Also:         IContainer > IIterable > IDict
# Also:         IContainer > IIterable > ISet

@interface.implementer(IBeforeTextAssignedEvent)
class BeforeTextAssignedEvent(BeforeSchemaFieldAssignedEvent):
    pass

@interface.implementer(IBeforeTextLineAssignedEvent)
class BeforeTextLineAssignedEvent(BeforeTextAssignedEvent):
    pass


@interface.implementer(IBeforeCollectionAssignedEvent)
class BeforeCollectionAssignedEvent(BeforeSchemaFieldAssignedEvent):
    object = None

@interface.implementer(IBeforeSequenceAssignedEvent)
class BeforeSequenceAssignedEvent(BeforeCollectionAssignedEvent):
    pass

@interface.implementer(IBeforeSetAssignedEvent)
class BeforeSetAssignedEvent(BeforeCollectionAssignedEvent):
    pass

@interface.implementer(IBeforeDictAssignedEvent)
class BeforeDictAssignedEvent(BeforeSchemaFieldAssignedEvent):
    pass

from zope.schema._field import BeforeObjectAssignedEvent
BeforeObjectAssignedEvent = BeforeObjectAssignedEvent

class InvalidValue(sch_interfaces.InvalidValue):
    """
    Adds a field specifically to carry the value that is invalid.
    """
    value = None

    def __init__(self, *args, **kwargs):
        super(InvalidValue, self).__init__(*args)
        if 'value' in kwargs:
            self.value = kwargs['value']
        if 'field' in kwargs:
            self.field = kwargs['field']

# And we monkey patch it in to InvalidValue as well
if not hasattr(sch_interfaces.InvalidValue, 'value'):
    setattr(sch_interfaces.InvalidValue, 'value', None)

# And give all validation errors a 'field'
if not hasattr(sch_interfaces.ValidationError, 'field'):
    setattr(sch_interfaces.ValidationError, 'field', None)

class IFromObject(interface.Interface):
    """
    Something that can convert one type of object to another,
    following validation rules (see :class:`zope.schema.interfaces.IFromUnicode`)
    """

    def fromObject(obj):
        """
        Attempt to convert the object to the required type following
        the rules of this object. Raises a TypeError or :class:`zope.schema.interfaces.ValidationError`
        if this isn't possible.
        """

class IVariant(sch_interfaces.IField, IFromObject):
    """
    Similar to :class:`zope.schema.interfaces.IObject`, but
    representing one of several different types.
    """

class IListOrTuple(sch_interfaces.IList):
    pass

def find_most_derived_interface(ext_self, iface_upper_bound, possibilities=None):
    """
    Search for the most derived version of the interface `iface_upper_bound`
    implemented by `ext_self` and return that. Always returns at least `iface_upper_bound`

    :keyword possibilities: An iterable of schemas to consider. If not given,
        all the interfaces provided by ``ext_self`` will be considered.
    """
    if possibilities is None:
        possibilities = interface.providedBy(ext_self)
    _iface = iface_upper_bound
    for iface in possibilities:
        if iface.isOrExtends(_iface):
            _iface = iface
    return _iface

try:
    from dm.zope.schema.interfaces import ISchemaConfigured as _ISchemaConfigured
except ImportError:
    _ISchemaConfigured = interface.Interface

class ISchemaConfigured(_ISchemaConfigured):
    """
    marker interface for ``SchemaConfigured`` classes.

    Used to facilitate the registration of forms and views.
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test support for schemas and interfaces, mostly in the form of Hamcrest matchers.

.. $Id$
"""

from __future__ import print_function, absolute_import, division

from zope.deferredimport import deprecatedFrom

deprecatedFrom("Matchers live in nti.testing.matchers",
               "nti.testing.matchers",
               "provides", "implements",
               "verifiably_provides", "validly_provides",
               "validated_by", "not_validated_by")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
For producing a JSON schema appropriate for use by clients, based on a Zope schema.

.. note:: This schema is ad-hoc and non-standard.

.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import interface

from zope.schema import interfaces as sch_interfaces
from zope.schema import vocabulary as sch_vocabulary

from nti.schema.interfaces import find_most_derived_interface

from six import string_types
from six import integer_types

#: Don't display this by default in the UI
TAG_HIDDEN_IN_UI = "nti.dataserver.users.field_hidden_in_ui"

#: Qualifying details about how the field should be treated, such as data source
TAG_UI_TYPE = 'nti.dataserver.users.field_type'

#: Overrides the value from the field itself
TAG_REQUIRED_IN_UI = 'nti.dataserver.users.field_required'

#: Overrides the value from the field itself, if true
TAG_READONLY_IN_UI = 'nti.dataserver.users.field_readonly'

# Known types
UI_TYPE_EMAIL = 'nti.dataserver.users.interfaces.EmailAddress'
UI_TYPE_HASHED_EMAIL = UI_TYPE_EMAIL + ":Hashed"  # So that a begins-with test will match either, making validation easier

#: Something that can be set once, typically during account creation
UI_TYPE_ONE_TIME_CHOICE = 'nti.dataserver.users.interfaces.OneTimeChoice'

def get_ui_type_from_interface(iface):
    ui_type = iface.getName()
    ui_type = ui_type[1:] if ui_type.startswith('I') else ui_type
    return ui_type
_interface_to_ui_type = interface_to_ui_type = get_ui_type_from_interface # BWC

def get_ui_type_from_field_interface(field):
    derived_field_iface = find_most_derived_interface(field, sch_interfaces.IField)
    if derived_field_iface is not sch_interfaces.IField:
        ui_type = get_ui_type_from_interface(derived_field_iface)
        return ui_type
    return None
_ui_type_from_field_iface = ui_type_from_field_iface = get_ui_type_from_field_interface # BWC

def get_ui_types_from_field(field):
    ui_type = ui_base_type = None
    _type = getattr(field, '_type', None)
    if isinstance(_type, type):
        ui_type = _type.__name__
    elif isinstance(_type, tuple):
        # Most commonly lists subclasses. Most commonly lists subclasses of strings
        if all((issubclass(x, string_types) for x in _type)):
            ui_type = 'basestring'
        elif all((issubclass(x, integer_types) for x in _type)):
            ui_type = 'int'
        elif all((issubclass(x, float) for x in _type)):
            ui_type = 'float'
    else:
        ui_type = get_ui_type_from_field_interface(field)

    if ui_type in ('unicode', 'str', 'basestring'):
        # These are all 'string' type

        # Can we be more specific?
        ui_type = get_ui_type_from_field_interface(field)
        if ui_type and ui_type not in ('TextLine', 'Text'): # pragma: no cover
            # Yes we can
            ui_base_type = 'string'
        else:
            ui_type = 'string'
            ui_base_type = 'string'
    elif ui_type in ('Number', 'float'):
        ui_base_type = 'float'
    elif ui_type in ('int', 'long'):
        ui_base_type = 'int'
    elif ui_type in ('bool',):
        ui_base_type = 'bool'
    return ui_type, ui_base_type

_ui_type_from_field = ui_type_from_field = get_ui_types_from_field # BWC

def get_data_from_choice_field(v, base_type=None):
    # Vocabulary could be a name or the vocabulary itself
    choices = ()
    vocabulary = None
    if sch_interfaces.IVocabulary.providedBy(v.vocabulary): # pragma: no cover
        vocabulary = v.vocabulary
    elif isinstance(v.vocabularyName, string_types):
        name = v.vocabularyName
        vocabulary = sch_vocabulary.getVocabularyRegistry().get(None, name)

    if vocabulary is not None:
        choices = []
        tokens = []
        for term in vocabulary:
            # For BWC, we do different things depending on whether
            # there is a title or not
            if getattr(term, 'title', None):
                try:
                    # like nti.externalization, but without the dependency
                    choice = term.toExternalObject()
                except AttributeError: # pragma: no cover
                    choice = {'token': term.token,
                              'value': term.value,
                              'title': term.title}

                choices.append(choice)
            else: # pragma: no cover
                choices.append(term.token)  # bare; ideally this would go away
            tokens.append(term.token)

        # common case, these will all be the same type
        if  not base_type \
            and all((isinstance(x, string_types) for x in tokens)):
            base_type = 'string'
    return choices, base_type
_process_choice_field = process_choice_field = get_data_from_choice_field

class JsonSchemafier(object):

    def __init__(self, schema, readonly_override=None):
        """
        Create a new schemafier.

        :param schema: The zope schema to use.
        :param bool readonly_override: If given, a boolean value that will replace all
            readonly values in the schema.
        """
        self.schema = schema
        self.readonly_override = readonly_override

    def _iter_names_and_descriptions(self):
        """
        Return an iterable across the names and descriptions of the schema.

        Subclass hook to change what is considered.
        """
        return self.schema.namesAndDescriptions(all=True)

    def allow_field(self, name, field):
        """
        Return if the field is allowed in the external schema
        """
        if field.queryTaggedValue(TAG_HIDDEN_IN_UI) or name.startswith('_'):
            return False
        return True

    def get_ui_types_from_field(self, field):
        """
        Return the type and base type for the specified field
        """
        return get_ui_types_from_field(field)
    ui_types_from_field = get_ui_types_from_field # BWC

    def get_data_from_choice_field(self, field, base_type=None):
        """
        Return the choices and base type for the specified field
        """
        return get_data_from_choice_field(field, base_type)
    process_choice_field = get_data_from_choice_field # BWC

    def post_process_field(self, name, field, item_schema):
        pass

    def make_schema(self):
        """
        Create the JSON schema.

        Individual fields of the schema will be checked and returned. See the various
        ``TAG`` constants for ways that the schema externalization can be influenced.

        :return: A dictionary consisting of dictionaries, one for each field. All the keys
            are strings and the values are strings, bools, numbers, or lists of primitives.
            Will be suitable for writing to JSON.
        """
        readonly_override = self.readonly_override

        ext_schema = {}
        for k, v in self._iter_names_and_descriptions():
            __traceback_info__ = k, v
            if interface.interfaces.IMethod.providedBy(v):
                continue
            # v could be a schema field or an interface.Attribute
            if not self.allow_field(k, v):
                continue

            required = getattr(v, 'required', None)
            required = v.queryTaggedValue(TAG_REQUIRED_IN_UI) or required

            if readonly_override is not None:
                readonly = readonly_override
            else:
                readonly = getattr(v, 'readonly', False)
                readonly = v.queryTaggedValue(TAG_READONLY_IN_UI) or readonly

            ui_base_type = None
            item_schema = {'name': k,
                           'required': required,
                           'readonly': readonly,
                           'min_length': getattr(v, 'min_length', None),
                           'max_length': getattr(v, 'max_length', None) }
            ui_type = v.queryTaggedValue(TAG_UI_TYPE)
            if not ui_type:
                ui_type, ui_base_type = self.get_ui_types_from_field(v)
            else:
                _, ui_base_type = self.get_ui_types_from_field(v)

            item_schema['type'] = ui_type
            item_schema['base_type'] = ui_base_type

            if sch_interfaces.IChoice.providedBy(v):
                choices, base_type = self.get_data_from_choice_field(v, ui_base_type)
                item_schema['choices'] = choices
                item_schema['base_type'] = base_type

            self.post_process_field(k, v, item_schema)
            ext_schema[k] = item_schema

        return ext_schema



#!/usr/bin/env python

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904


from hamcrest import is_not
from hamcrest import contains

from hamcrest import has_length
from hamcrest import assert_that
from hamcrest import has_property
from hamcrest import calling
from hamcrest import raises
does_not = is_not

import unittest

from zope import interface
from . import IUnicode

from nti.schema.field import Number
from nti.schema.field import Object


from nti.schema.field import DictFromObject
from nti.schema.field import ValidTextLine as TextLine

from ..schema import PermissiveSchemaConfigured

from nti.schema.interfaces import IBeforeDictAssignedEvent


class TestMisc(unittest.TestCase):

    def test_fixup_name(self):
        from zope.schema.fieldproperty import FieldPropertyStoredThroughField

        field = Object(IUnicode)
        field.__name__ = 'field'

        field_property = FieldPropertyStoredThroughField(field)
        field = field_property.field

        assert_that(field, has_property('__name__', '__st_field_st'))
        assert_that(field, has_property('__fixup_name__', 'field'))

class TestSchemaConfigured(unittest.TestCase):

    def test_init(self):
        class IA(interface.Interface):
            field = Number()

        @interface.implementer(IA)
        class A(PermissiveSchemaConfigured):
            pass

        # All of these are valid ways to call it
        A()
        A(thing='abc')
        a = A(field=1)
        assert_that(a, has_property('field', 1))

        # But we can flip off the extra
        A.SC_PERMISSIVE = False
        A()
        assert_that(calling(A).with_args(thing=1),
                    raises(TypeError, 'non schema keyword'))
        a = A(field=1)
        assert_that(a, has_property('field', 1))


from . import SchemaLayer
from zope.component import eventtesting

class TestConfigured(unittest.TestCase):

    layer = SchemaLayer


    def test_dict(self):

        dict_field = DictFromObject(key_type=TextLine(), value_type=Number())
        dict_field.__name__ = 'dict'

        class X(object):
            pass

        x = X()
        dict_field.set(x, dict_field.fromObject({'k': '1'}))

        assert_that(x, has_property('dict', {'k': 1.0}))

        events = eventtesting.getEvents(IBeforeDictAssignedEvent)
        assert_that(events, has_length(1))
        assert_that(events, contains(has_property('object', {'k': 1.0})))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Helpers for hashing and equality based on a list of names.

.. $Id$
"""

from __future__ import print_function,  absolute_import, division
__docformat__ = "restructuredtext en"


def _superhash(value):
    """
    Returns something that's hashable, either the value itself,
    or a tuple that can in turn be hashed
    """
    try:
        # By returning the original value, if it was hashable,
        # we may get better ultimate hash results;
        # the cost is hashing that value twice
        hash(value)
    except TypeError:
        # Dict?
        try:
            # Sort these, they have no order
            items = sorted(value.items())
        except AttributeError:
            # mutable iterable, which we must not sort
            items = value

        return tuple([_superhash(item)
                      for item
                      in items])
    else:
        return value

def EqHash(*names,
           **kwargs):
    """
    A class decorator factory for the common pattern of writing
    ``__eq__``/``__ne__`` and ``__hash__`` methods that check the same
    list of attributes on a given object.

    Right now, you must pass as individual arguments the property
    names to check; in the future, you may be able to pass a schema
    interface that defines the property names. Property names are compared
    for equality in the order they are given, so place the cheapest first.

    Additional parameters are only available via keywords::

      @EqHash('a', 'b')
      class Thing(object):
         a = 'a'
         b = 'b'

      @EqHash('c', include_super=True)
      class ChildThing(Thing):
         c = 'c'

    :keyword include_super: If set to ``True`` (*not* the default)
        then the equality (and perhaps hash) values of super will be considered.
    :keyword superhash: If set to ``True`` (*not* the default),
        then the hash function will be made to support certain
        mutable types (lists and dictionaries) that ordinarily cannot
        be hashed. Use this only when those items are functionally
        treated as immutable.
    :keyword include_type: If set to ``True`` (*not* the default),
        equality will only be true if the other object is an instance
        of the class this is declared on. Use this only when there are
        a series of subclasses who differ in no attributes but should not
        compare equal to each other. Note that this can lead to violating
        the commutative property.

    """

    _include_super = kwargs.pop('include_super', False)
    superhash = kwargs.pop("superhash", False)
    _include_type = kwargs.pop('include_type', False)

    if kwargs:
        raise TypeError("Unexpected keyword args", kwargs)
    if not names and not _include_super and not _include_type:
        raise TypeError("Asking to hash/eq nothing, but not including super or type")


    def x(cls):
        __eq__, __hash__, __ne__ = _eq_hash(cls, names,
                                            _include_super, _include_type, superhash)
        cls.__eq__ = __eq__
        cls.__hash__ = __hash__
        cls.__ne__ = __ne__
        return cls
    return x

def _eq_hash(cls, names, include_super, include_type, superhash): # pylint:disable=I0011,W0622,R0912
    names = tuple((str(x) for x in names)) # make sure they're native strings, not unicode on Py2
    # We assume the class hierarchy of these objects does not change
    if include_super:
        superclass = cls.__mro__[1]
        superclass_eq = superclass.__eq__
        superclass_hash = superclass.__hash__

    # 1 and 0 are constants and faster to load than the globals True/False
    # (in python 2)

    def __eq__(self, other):
        if self is other:
            return 1

        if include_type:
            if not isinstance(other, cls):
                return 0

        if include_super:
            s = superclass_eq(self, other)
            if s is NotImplemented or not s:
                return s


        # We take these one at a time (rather than using
        # operator.attrgetter). In the cases where some attributes
        # are computed, this can be more efficient if we discover
        # a mismatch early. Also, it lets us easily distinguish
        # between an AttributeError on self (which is a
        # programming error in calling EqHash) or the other object
        _ga = getattr # load_fast in the loop
        for name in names:
            my_val = _ga(self, name)
            try:
                other_val =_ga(other, name)
            except AttributeError:
                return NotImplemented
            else:
                # Amusingly, even though this does
                # more push/pop than inlining in the
                # body of the try block, it benchmarks
                # faster
                if my_val != other_val:
                    return 0
        return 1

    def __ne__(self, other):
        eq = __eq__(self, other)
        if eq is NotImplemented:
            return eq
        return not eq

    # Our contract for include_super says that hashing
    # may or may not be included. It shouldn't affect the results
    # if we do not actually include it, unless there are no values
    # being hashed from this object. However, for consistency,
    # we always include it if asked
    seed = hash(names)

    if superhash:
        def _hash(values):
            # We're hashing the sequence of superhash values
            # for each value; we could probably do better?
            return hash( tuple([_superhash(x) for x in values]) )
    else:
        def _hash(values):
            return hash(tuple(values))

    def __hash__(self):
        h = seed
        if include_super:
            h ^= superclass_hash(self) << 2

        # If we or-equal for every attribute separately, we
        # easily run the risk of saturating the integer. So we boil
        # all attributes down to one value to hash
        if names:
            ga = getattr
            h ^= _hash( [ga(self, name) for name in names] )
        return h

    return __eq__, __hash__, __ne__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Event handlers.

.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import component

from zope.component import handle

from nti.schema.interfaces import IBeforeSchemaFieldAssignedEvent

@component.adapter(IBeforeSchemaFieldAssignedEvent)
def before_object_assigned_event_dispatcher(event):
    """
    Listens for :mod:`zope.schema` fields to fire :class:`~.IBeforeSchemaFieldAssignedEvent`,
    and re-dispatches these events based on the value being assigned, the object being assigned to,
    and of course the event (note that :class:`~zope.schema.interfaces.IBeforeObjectAssignedEvent` is a
    sub-interface of :class:`.~IBeforeSchemaFieldAssignedEvent`).

    This is analogous to :func:`zope.component.event.objectEventNotify`
    """
    handle( event.object, event.context, event )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Schema fields.

These fields produce better validation errors than the standard
:mod:`zope.schema` fields do. All the standard fields are also aliased
to be imported from this module.

.. TODO: This module is big enough it should be factored into a package and sub-modules.
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from nti.schema import MessageFactory as _

import re
from six import string_types
from six import reraise
import sys

from zope.schema import Object as _ObjectBase

from .schema import SchemaConfigured
SchemaConfigured = SchemaConfigured

import zope.interface.common.idatetime

from zope import schema
from zope import interface

from zope.event import notify
from zope.schema import interfaces as sch_interfaces

# Re-export some things as part of our public API so we can
# later re-implement them locally if needed
from zope.schema import Bool
Bool = Bool
from zope.schema import Date
Date = Date
from zope.schema import Datetime
Datetime = Datetime
DateTime = Datetime
from zope.schema import Decimal
Decimal = Decimal
from zope.schema import Dict
Dict = Dict
from zope.schema import List
List = List
from zope.schema import Text
Text = Text
from zope.schema import TextLine
TextLine = TextLine
from zope.schema import Timedelta
Timedelta = Timedelta
from zope.schema import Choice
Choice = Choice
from zope.schema import Tuple
Tuple = Tuple
from zope.schema import FrozenSet
FrozenSet = FrozenSet
from zope.schema import Set
Set = Set
from zope.schema import Iterable
Iterable = Iterable

from nti.schema.interfaces import IVariant
from nti.schema.interfaces import IFromObject
from nti.schema.interfaces import IListOrTuple

import numbers
import collections

from nti.schema.interfaces import BeforeSetAssignedEvent
from nti.schema.interfaces import BeforeDictAssignedEvent
from nti.schema.interfaces import BeforeTextAssignedEvent
from nti.schema.interfaces import BeforeObjectAssignedEvent
from nti.schema.interfaces import BeforeTextLineAssignedEvent
from nti.schema.interfaces import BeforeSequenceAssignedEvent
from nti.schema.interfaces import BeforeSchemaFieldAssignedEvent

def _do_set(self, context, value, cls, factory):
    try:
        event = factory(value, self.__name__, context)
        notify(event)
        value = event.object
        super(cls, self).set(context, value)
    except sch_interfaces.ValidationError as e: # pragma: no cover
        # This shouldn't happen, set() doesn't typically validate.
        self._reraise_validation_error(e, value)

def __make_set(cls, eventfactory):

    def set(self, context, value):
        return _do_set(self, context, value, cls, eventfactory)

    cls.set = set

def __with_set(eventfactory=BeforeSchemaFieldAssignedEvent):
    def X(cls):
        __make_set(cls, eventfactory)
        return cls
    return X

from nti.schema.interfaces import InvalidValue
InvalidValue = InvalidValue

class FieldValidationMixin(object):
    """
    A field mixin that causes slightly better errors to be created.
    """

    @property
    def __fixup_name__(self):
        """
        The :class:`zope.schema.fieldproperty.FieldPropertyStoredThroughField` class mangles
        the field name; this undoes that mangling.
        """
        if self.__name__ and self.__name__.startswith('__st_') and self.__name__.endswith('_st'):
            return self.__name__[5:-3]
        return self.__name__

    def _fixup_validation_error_args(self, e, value):
        # Called when the exception has one argument, which is usually, though not always,
        # the message
        e.args = (value, e.args[0], self.__fixup_name__)

    def _fixup_validation_error_no_args(self, e, value):
        # Called when there are no arguments
        e.args = (value, str(e), self.__fixup_name__)

    def _reraise_validation_error(self, e, value, _raise=False):
        if len(e.args) == 1:  # typically the message is the only thing
            self._fixup_validation_error_args(e, value)
        elif len(e.args) == 0:  # Typically a SchemaNotProvided. Grr.
            self._fixup_validation_error_no_args(e, value)
        elif isinstance(e, sch_interfaces.TooShort) and len(e.args) == 2:
            # Note we're capitalizing the field in the message.
            e.i18n_message = _('${field} is too short.',
                                mapping={'field': self.__fixup_name__.capitalize(),
                                        'minLength': e.args[1]})
            e.args = (self.__fixup_name__.capitalize() + ' is too short.',
                       self.__fixup_name__,
                       value)
        elif isinstance(e, sch_interfaces.TooLong) and len(e.args) == 2:
            e.i18n_message = _('${field} is too long. ${max_size} character limit.',
                                mapping={'field': self.__fixup_name__.capitalize(),
                                        'max_size': e.args[1]})
            e.args = (self.__fixup_name__.capitalize() + ' is too long.',
                       self.__fixup_name__,
                       value)
        e.field = self
        if not getattr(e, 'value', None):
            e.value = value
        if _raise:
            raise e
        raise

    def _validate(self, value):
        try:
            super(FieldValidationMixin, self)._validate(value)
        except sch_interfaces.WrongContainedType as e:
            # args[0] will either be a list of Exceptions or a list of tuples, (name, exception),
            # depending who did the validating (dm.zope.schema doing the later)
            e.errors = [arg[1] if isinstance(arg, tuple) else arg for arg in e.args[0]]
            e.value = value
            e.field = self
            raise
        except sch_interfaces.ValidationError as e:
            self._reraise_validation_error(e, value)

@interface.implementer(sch_interfaces.IObject)
class ValidDatetime(FieldValidationMixin, Datetime):
    """
    Unlike the standard datetime, this will check that the
    given object is an instance of IDatetime, and raise
    the same error as object does.
    """

    schema = zope.interface.common.idatetime.IDateTime

    def _validate(self, value):
        try:
            super(ValidDatetime, self)._validate(value)
        except sch_interfaces.WrongType as e:
            raise sch_interfaces.SchemaNotProvided(value, e.__doc__, self.__fixup_name__, self.schema, list(interface.providedBy(value)))

        # schema has to be provided by value
        if not self.schema.providedBy(value):  # pragma: no cover
            raise sch_interfaces.SchemaNotProvided

if hasattr(dict, 'iteritems'):
    _iteritems = dict.iteritems
else:
    _iteritems = dict.items

class Object(FieldValidationMixin, _ObjectBase):
    """
    Improved ``zope.schema.Object``.
    """

    # The original dm.zope.schema.Object that we were using before
    # contains extra validation logic, but as far as I can tell, it's
    # pretty much a duplicate of zope.schema.Object. Perhaps that wasn't
    # always the case? At any rate, elide it, because it never seems to get
    # called (the error would already have been raised.)

    def _fixup_validation_error_no_args(self, e, value):
        e.args = (value, e.__doc__, self.__fixup_name__, self.schema, list(interface.providedBy(value)))

@interface.implementer(IVariant)
class Variant(FieldValidationMixin, schema.Field):
    """
    Similar to :class:`zope.schema.Object`, but accepts one of many non-overlapping
    interfaces.
    """

    fields = ()

    def __init__(self, fields, variant_raise_when_schema_provided=False, **kwargs):
        """
        :param fields: A list or tuple of field instances.
        :keyword variant_raise_when_schema_provided: If ``True``, then
            if a value is provided to ``validate`` that implements
            the schema of a particular field, and that field raised
            a validation error, that error will be propagated instead
            of the error raised by the last field, and no additional fields
            will be asked to do validation.

        """
        if not fields or not all((sch_interfaces.IField.providedBy(x) for x in fields)):
            raise sch_interfaces.WrongType()

        # assign our children first so anything we copy to them as a result of the super
        # constructor (__name__) gets set
        self.fields = list(fields)
        for f in self.fields:
            f.__parent__ = self

        self._raise_when_provided = variant_raise_when_schema_provided
        super(Variant, self).__init__(**kwargs)

    def __get_name(self):
        return self.__dict__.get('__name__', '')

    def __set_name(self, name):
        self.__dict__['__name__'] = name
        for field in self.fields:
            field.__name__ = name
    __name__ = property(__get_name, __set_name)

    def getDoc(self):
        doc = super(Variant, self).getDoc()
        doc += '\n\nValue is one of:'
        for field in self.fields:
            fielddoc = field.getDoc()
            if not fielddoc:
                fielddoc = getattr(type(field), '__doc__', '')
            if fielddoc:
                doc += '\n\n\t' + fielddoc
        # Definition lists must end with a blank line
        doc += '\n'
        return doc

    def bind(self, obj):
        clone = super(Variant, self).bind(obj)
        clone.fields = [x.bind(obj) for x in clone.fields]
        for f in clone.fields:
            f.__parent__ = clone
        return clone

    def _validate(self, value):
        super(Variant, self)._validate(value)
        for field in self.fields:
            try:
                field.validate(value)
                # one of them accepted, yay!
                return
            except sch_interfaces.ValidationError as e:
                if self._raise_when_provided and hasattr(field, 'schema') and field.schema.providedBy(value):
                    raise
        # We get here only by all of them throwing an exception.
        # we re-raise the last thing thrown
        self._reraise_validation_error(e, value)

    def fromObject(self, obj):
        """
        Similar to `fromUnicode`, attempts to turn the given object into something
        acceptable and valid for this field. Raises a TypeError, ValueError, or
        schema ValidationError if this cannot be done. Adaptation is attempted in the order
        in which fields were given to the constructor. Some fields cannot be used to adapt.
        """

        exc_info = None

        for field in self.fields:
            try:
                # Three possible ways to convert: adapting the schema of an IObject,
                # using a nested field that is IFromObject, or an IFromUnicode if the object
                # is a string.

                converter = None
                # Most common to least common
                if sch_interfaces.IObject.providedBy(field):
                    converter = field.schema
                elif sch_interfaces.IFromUnicode.providedBy(field) and isinstance(obj, string_types):
                    converter = field.fromUnicode
                elif IFromObject.providedBy(field):
                    converter = field.fromObject

                # Try to convert and validate
                adapted = converter(obj)
            except (TypeError, sch_interfaces.ValidationError):
                # Nope, no good
                exc_info = sys.exc_info()
            else:
                # We got one that like the type. Do the validation
                # now, and then return. Don't try to convert with others;
                # this is probably our best error
                try:
                    field.validate(adapted)
                    return adapted
                except sch_interfaces.SchemaNotProvided:
                    # Except in one case. Some schema provides adapt to something
                    # that they do not actually want (e.g., ISanitizedHTMLContent can adapt as IPlainText when empty)
                    # so ignore that and keep trying
                    exc_info = sys.exc_info()

        # We get here if nothing worked and re-raise the last exception
        try:
            reraise(*exc_info)
        finally:
            del exc_info

    def set(self, context, value):
        # Try to determine the most appropriate event to fire
        # Order matters. It would kind of be nice to direct this to the appropriate
        # field itself, but that's sort of hard.
        types = ((string_types, BeforeTextAssignedEvent),
                 (collections.Mapping, BeforeDictAssignedEvent),
                 (collections.Sequence, BeforeSequenceAssignedEvent),
                 (object, BeforeObjectAssignedEvent))
        for kind, factory in types:
            if isinstance(value, kind):
                _do_set(self, context, value, Variant, factory)
                return

class ObjectLen(FieldValidationMixin, schema.MinMaxLen, _ObjectBase):  # order matters
    """
    Allows specifying a length for arbitrary object fields (though the
    objects themselves must support the `len` function.
    """

    def __init__(self, sch, min_length=0, max_length=None, **kwargs):
        # match the calling sequence of Object, which uses a non-keyword
        # argument for schema.
        # But to work with the superclass, we have to pass it as a keyword arg.
        # it's weird.
        super(ObjectLen, self).__init__(schema=sch, min_length=min_length, max_length=max_length, **kwargs)

    def _fixup_validation_error_no_args(self, e, value):
        e.args = (value, e.__doc__, self.__fixup_name__, self.schema, list(interface.providedBy(value)))

class Int(FieldValidationMixin, schema.Int):

    def fromUnicode(self, value):
        # Allow empty strings
        result = super(Int, self).fromUnicode(value) if value else None
        return result

class Float(FieldValidationMixin, schema.Float):

    def fromUnicode(self, value):
        result = super(Float, self).fromUnicode(value) if value else None
        return result

class Number(FieldValidationMixin, schema.Float):
    """
    A field that parses like a float from a string, but accepts any number.
    """
    _type = numbers.Number

@__with_set()
class ValidChoice(FieldValidationMixin, schema.Choice):
    pass

@__with_set()
class ValidBytesLine(FieldValidationMixin, schema.BytesLine):
    pass

@__with_set()
class ValidBytes(FieldValidationMixin, schema.Bytes):
    pass

@__with_set(BeforeTextAssignedEvent)
class ValidText(FieldValidationMixin, schema.Text):
    """
    A text line that produces slightly better error messages. They will all
    have the 'field' property.

    We also fire :class:`IBeforeTextAssignedEvent`, which the normal
    mechanism does not.
    """

@__with_set(BeforeTextLineAssignedEvent)
class ValidTextLine(FieldValidationMixin, schema.TextLine):
    """
    A text line that produces slightly better error messages. They will all
    have the 'field' property.

    We also fire :class:`IBeforeTextLineAssignedEvent`, which the normal
    mechanism does not.
    """

class DecodingValidTextLine(ValidTextLine):
    """
    A text type that will attempt to decode non-unicode
    data as UTF-8.

    This primarily exists for legacy support (tests and persisted data).
    """

    def validate(self, value):
        if isinstance(value, bytes):
            value = value.decode('utf-8')  # let raise UnicodeDecodeError
        super(DecodingValidTextLine, self).validate(value)
        return value # tests

    # fromUnicode calls validate, so no need to duplicate

class ValidRegularExpression(ValidTextLine):

    def __init__(self, pattern, flags=re.U | re.I | re.M, *args, **kwargs):
        super(ValidRegularExpression, self).__init__(*args, **kwargs)
        self.flags = flags
        self.pattern = pattern
        self.prog = re.compile(pattern, flags)

    def constraint(self, value):
        return self.prog.match(value) is not None

ValidRegEx = ValidRegularExpression

class ValidURI(FieldValidationMixin, schema.URI):

    def _fixup_validation_error_args(self, e, value):
        if isinstance(e, sch_interfaces.InvalidURI):
            # This class differs by using the value as the argument, not
            # a message
            e.__doc__ = e.__doc__.replace('URI', 'URL')
            e.args = (value, e.__doc__, self.__fixup_name__)
            e.message = e.i18n_message = e.__doc__
        else: # pragma: no cover
            super(ValidURI, self)._fixup_validation_error_args(e, value)

class HTTPURL(ValidURI):
    """
    A URI field that ensures and requires its value to be an absolute
    HTTP/S URL.
    """

    def fromUnicode(self, value):
        # This can wind up producing something invalid if an
        # absolute URI was already given for mailto: for whatever.
        # None of the regexs (zopes or grubers) flag that as invalid.
        # so we try to
        orig_value = value
        if value:
            lower = value.lower()
            if not lower.startswith('http://') and not lower.startswith('https://'):
                # assume http
                value = 'http://' + value
        result = super(HTTPURL, self).fromUnicode(value)
        if result.count(':') != 1:
            self._reraise_validation_error(sch_interfaces.InvalidURI(orig_value), orig_value, _raise=True)

        return result

class _ValueTypeAddingDocMixin(object):
    """
    A mixin for fields that wrap a value type field (e.g., Object)
    to copy the nested documentation to the parent so it is visible
    in :mod:`repoze.sphinx.autointerface`.
    """

    document_value_type = True

    def getDoc(self):
        doc = super(_ValueTypeAddingDocMixin, self).getDoc()
        if self.document_value_type:
            value_type = getattr(self, 'value_type', None)
            if value_type is not None:
                doc += '\nThe value type is documented as:\n\t' + value_type.getDoc()
                doc += '\n'
            _type = getattr(self, 'accept_types', getattr(self, '_type', None))
            def _class_dir(t):
                mod = t.__module__ + '.' if t.__module__ and t.__module__ != '__builtin__' else ''
                return ':class:`' + mod + t.__name__ + '`'

            if isinstance(_type, type): # pragma: no cover
                doc += '\nThe acceptable class is ' + _class_dir(_type) + '.'
            elif _type:
                types = [_class_dir(t) for t in _type]
                doc += '\nThe acceptable classes are ' + ' , '.join(types) + '.'
        return doc

@__with_set(BeforeSequenceAssignedEvent)
class IndexedIterable(_ValueTypeAddingDocMixin, FieldValidationMixin, schema.List):
    """
    An arbitrary (indexable) iterable, not necessarily a list or tuple;
    either of those would be acceptable at any time (however, so would a string,
    so be careful. Try ListOrTuple if that's a problem).

    The values may be homogeneous by setting the value_type.
    """
    _type = None  # Override from super to not force a list

@interface.implementer(IListOrTuple)
class ListOrTuple(IndexedIterable):
    _type = (list, tuple)

class _SequenceFromObjectMixin(object):
    accept_types = None
    _default_type = list

    def _converter_for(self, field):
        if hasattr(field, 'fromObject'):
            converter = field.fromObject
        elif hasattr(field, 'fromUnicode'):  # here's hoping the values are strings
            converter = field.fromUnicode
        return converter

    def _do_fromObject(self, context):
        converter = self._converter_for(self.value_type)
        result = [converter(x) for x in context]
        return result

    def fromObject(self, context):
        check_type = self.accept_types or self._type
        if check_type is not None and not isinstance(context, check_type):
            raise sch_interfaces.WrongType(context, self._type)

        result = self._do_fromObject(context)
        if isinstance(self._type, type) and self._type is not self._default_type:  # single type is a factory
            result = self._type(result)
        return result

@interface.implementer(IFromObject)
class ListOrTupleFromObject(_SequenceFromObjectMixin, ListOrTuple):
    """
    The field_type MUST be a :class:`Variant`, or more generally,
    something supporting :class:`IFromObject` or :class:`IFromUnicode`
    """

    def __init__(self, *args, **kwargs):
        super(ListOrTupleFromObject, self).__init__(*args, **kwargs)
        if not IFromObject.providedBy(self.value_type):
            raise sch_interfaces.WrongType()

@interface.implementer(IFromObject)
class TupleFromObject(_ValueTypeAddingDocMixin, _SequenceFromObjectMixin, FieldValidationMixin, schema.Tuple):
    """
    The field_type MUST be a :class:`Variant`, or more generally,
    something supporting :class:`IFromObject`. When setting through this object,
    we will automatically convert lists and only lists to tuples (for convenience coming
    in through JSON)
    """
    accept_types = (list, tuple)
    def set(self, context, value):
        if isinstance(value, list):
            value = tuple(value)

        _do_set(self, context, value, TupleFromObject, BeforeSequenceAssignedEvent)

    def validate(self, value):
        if isinstance(value, list):
            value = tuple(value)
        super(TupleFromObject, self).validate(value)

@interface.implementer(IFromObject)
@__with_set(BeforeDictAssignedEvent)
class DictFromObject(_ValueTypeAddingDocMixin,
                     _SequenceFromObjectMixin,
                     FieldValidationMixin,
                     schema.Dict):
    """
    The `key_type` and `value_type` must be supporting :class:`IFromObject` or :class:`.IFromUnicode`.
    """

    def _do_fromObject(self, context):
        key_converter = self._converter_for(self.key_type)
        value_converter = self._converter_for(self.value_type)
        return {key_converter(k): value_converter(v) for k, v in _iteritems(context)}

@__with_set(BeforeSetAssignedEvent)
class ValidSet(_ValueTypeAddingDocMixin, FieldValidationMixin, schema.Set):
    pass

class UniqueIterable(ValidSet):
    """
    An arbitrary iterable, not necessarily an actual :class:`set` object,
    but one whose contents are unique. Use this when you can
    return a :class:`set`, :class:`frozenset` or empty tuple. These should be
    sequences that suport the ``in`` operator.
    """
    _type = None  # Override to not force a set

    def __init__(self, *args, **kwargs):
        # If they do not specify a min_length in the arguments,
        # then change it to None. This way we are compatible with
        # a generator value. Superclass specifies both a class value
        # and a default argument
        no_min_length = False
        if 'min_length' not in kwargs:
            no_min_length = True

        super(UniqueIterable, self).__init__(*args, **kwargs)
        if no_min_length:
            self.min_length = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""


.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904

import unittest
from hamcrest import assert_that
from hamcrest import is_
from hamcrest import calling
from hamcrest import raises
from hamcrest import has_property
from hamcrest import contains
from hamcrest import has_length
from hamcrest import none

# use old imports to test bwc
from nti.schema.testing import validated_by
from nti.schema.testing import not_validated_by
from nti.schema.testing import verifiably_provides


from . import SchemaLayer
from zope.component import eventtesting

from nti.schema.field import DecodingValidTextLine
from nti.schema.field import FieldValidationMixin
from nti.schema.field import Float
from nti.schema.field import HTTPURL
from nti.schema.field import IndexedIterable
from nti.schema.field import Int
from nti.schema.field import ListOrTuple
from nti.schema.field import ListOrTupleFromObject
from nti.schema.field import Number
from nti.schema.field import Object
from nti.schema.field import ObjectLen
from nti.schema.field import TupleFromObject
from nti.schema.field import UniqueIterable
from nti.schema.field import ValidDatetime
from nti.schema.field import ValidRegularExpression
from nti.schema.field import ValidTextLine as TextLine
from nti.schema.field import Variant

from zope.schema import Dict

from nti.schema.interfaces import IVariant
from nti.schema.interfaces import IBeforeDictAssignedEvent
from nti.schema.interfaces import IBeforeSequenceAssignedEvent

from . import IUnicode

from zope.schema.interfaces import SchemaNotProvided
from zope.schema.interfaces import TooLong
from zope.schema.interfaces import TooShort
from zope.schema.interfaces import InvalidURI
from zope.schema.interfaces import WrongType
from zope.interface.common import interfaces as cmn_interfaces


class TestObjectLen(unittest.TestCase):


    def test_objectlen(self):
        # If we have the inheritance messed up, we will have problems
        # creating, or we will have problems validating one part or the other.

        olen = ObjectLen(IUnicode, max_length=5)  # default val for min_length

        olen.validate('a')
        olen.validate('')

        assert_that(calling(olen.validate).with_args(object()),
                     raises(SchemaNotProvided))

        assert_that(calling(olen.validate).with_args('abcdef'),
                     raises(TooLong))

    def test_objectlen_short(self):
        olen = ObjectLen(IUnicode, min_length=5)

        assert_that(calling(olen.validate).with_args('abc'),
                     raises(TooShort))

class TestHTTPUrl(unittest.TestCase):

    def test_http_url(self):

        http = HTTPURL(__name__='foo')

        assert_that(http.fromUnicode('www.google.com'),
                     is_('http://www.google.com'))

        assert_that(http.fromUnicode('https://www.yahoo.com'),
                     is_('https://www.yahoo.com'))

        try:
            http.fromUnicode('mailto:jason@nextthought.com')
            self.fail("Must raise")
        except InvalidURI as ex:
            exception = ex

        assert_that(exception, has_property('field', http))
        assert_that(exception, has_property('value', 'mailto:jason@nextthought.com'))
        assert_that(exception, has_property('message', 'The specified URL is not valid.'))

class TestVariant(unittest.TestCase):

    def test_variant(self):

        syntax_or_lookup = Variant((Object(cmn_interfaces.ISyntaxError),
                                    Object(cmn_interfaces.ILookupError),
                                    Object(IUnicode)))

        assert_that(syntax_or_lookup, verifiably_provides(IVariant))

        # validates
        assert_that(SyntaxError(), validated_by(syntax_or_lookup))
        assert_that(LookupError(), validated_by(syntax_or_lookup))

        # doesn't validate
        assert_that(b'foo', not_validated_by(syntax_or_lookup))

        assert_that(syntax_or_lookup.fromObject('foo'), is_('foo'))

        assert_that(calling(syntax_or_lookup.fromObject).with_args(object()),
                     raises(TypeError))

        # cover
        syntax_or_lookup.getDoc()

    def test_complex_variant(self):

        dict_field = Dict(key_type=TextLine(), value_type=TextLine())
        string_field = Object(IUnicode)
        list_of_numbers_field = ListOrTuple(value_type=Number())

        variant = Variant((dict_field, string_field, list_of_numbers_field))
        variant.getDoc()  # cover
        # It takes all these things
        for d in { 'k': 'v'}, 'foo', [1, 2, 3]:
            assert_that(d, validated_by(variant))

        # It rejects these
        for d in {'k': 1}, b'foo', [1, 2, 'b']:
            assert_that(d, not_validated_by(variant))

        # A name set now is reflected down the line
        variant.__name__ = 'baz'
        for f in variant.fields:
            assert_that(f, has_property('__name__', 'baz'))

        # and in clones
        clone = variant.bind(object())
        for f in clone.fields:
            assert_that(f, has_property('__name__', 'baz'))

        # which doesn't change the original
        clone.__name__ = 'biz'
        for f in clone.fields:
            assert_that(f, has_property('__name__', 'biz'))

        for f in variant.fields:
            assert_that(f, has_property('__name__', 'baz'))

        # new objects work too
        new = Variant(variant.fields, __name__='boo')
        for f in new.fields:
            assert_that(f, has_property('__name__', 'boo'))

    def test_variant_from_object(self):
        field = Variant((TupleFromObject(HTTPURL()),))

        res = field.fromObject(['http://example.com'])
        assert_that(res, is_(('http://example.com',)))

    def test_converts_but_not_valid(self):
        # If the schema accepts the input, but the validation refuses,
        # keep going.
        class WeirdField(Object):
            schema = IUnicode

            def validate(self, value):
                raise SchemaNotProvided()
        weird_field = WeirdField(IUnicode)
        accept_field = Number()

        field = Variant((weird_field, accept_field),
                        variant_raise_when_schema_provided=True)
        assert_that(field.fromObject("1.0"),
                    is_(1.0))

        assert_that(calling(field.validate).with_args('1.0'),
                    raises(SchemaNotProvided))

    def test_invalid_construct(self):
        assert_that(calling(Variant).with_args(()),
                    raises(WrongType))

class TestConfiguredVariant(unittest.TestCase):

    layer = SchemaLayer

    def test_nested_variants(self):
        # Use case: Chat messages are either a Dict, or a N
        # ote-like body, which itself is a list of variants

        dict_field = Dict(key_type=TextLine(), value_type=TextLine())

        string_field = Object(IUnicode)
        number_field = Number()
        list_of_strings_or_numbers = ListOrTuple(value_type=Variant((string_field, number_field)))

        assert_that([1, '2'], validated_by(list_of_strings_or_numbers))
        assert_that({'k': 'v'}, validated_by(dict_field))

        dict_or_list = Variant((dict_field, list_of_strings_or_numbers))

        assert_that([1, '2'], validated_by(dict_or_list))
        assert_that({'k': 'v'}, validated_by(dict_or_list))

        class X(object):
            pass

        x = X()
        dict_or_list.set(x, [1, '2'])

        events = eventtesting.getEvents(IBeforeSequenceAssignedEvent)
        assert_that(events, has_length(1))
        assert_that(events, contains(has_property('object', [1, '2'])))

        eventtesting.clearEvents()

        dict_or_list.set(x, {'k': 'v'})
        events = eventtesting.getEvents(IBeforeDictAssignedEvent)
        assert_that(events, has_length(1))
        assert_that(events, contains(has_property('object', {'k': 'v'})))

class TestUniqueIterable(unittest.TestCase):

    def test_min_length(self):
        field = UniqueIterable(__name__='foo')
        assert_that(field, has_property('min_length', none()))

        class Thing(object):
            foo = None
        thing = Thing()
        field.set(thing, ())

        assert_that(thing, has_property('foo', ()))

class TestTupleFromObject(unittest.TestCase):

    def test_set(self):
        field = TupleFromObject(__name__='foo')

        class Thing(object):
            foo = None

        thing = Thing()
        field.validate([1, 2])
        field.set(thing, [1, 2])
        assert_that(thing, has_property('foo', (1, 2)))

        # But arbitrary iterables not validated...
        assert_that(calling(field.validate).with_args('abc'),
                    raises(WrongType))

        # Although they can be set...
        field.set(thing, 'abc')

    def test_wrong_type_from_object(self):
        field = TupleFromObject()
        assert_that(calling(field.fromObject).with_args('abc'),
                    raises(WrongType))

    def test_valid_type_from_object_unicode(self):
        field = TupleFromObject(HTTPURL())
        res = field.fromObject(['http://example.com'])
        assert_that(res, is_(('http://example.com',)))

    def test_valid_type_from_object_object(self):
        # Nested layers of fromObject and fromUnicode
        field = TupleFromObject(Variant((HTTPURL(),)))
        res = field.fromObject(['http://example.com'])
        assert_that(res, is_(('http://example.com',)))

class TestListOrTupleFromObject(unittest.TestCase):

    def test_invalid_construct(self):
        assert_that(calling(ListOrTupleFromObject).with_args(Object(IUnicode)),
                    raises(WrongType))

class TestIndexedIterable(unittest.TestCase):

    def test_accepts_str(self):
        field = IndexedIterable(__name__='foo')
        class Thing(object):
            foo = None

        thing = Thing()
        field.set(thing, 'abc')
        assert_that(thing, has_property('foo', 'abc'))

class TestDecodingValidTextLine(unittest.TestCase):

    def test_decode(self):
        field = DecodingValidTextLine()
        res = field.validate(b'abc')
        assert_that(res, is_(u'abc'))

class TestNumber(unittest.TestCase):

    def test_allow_empty(self):
        assert_that(Float().fromUnicode(''), is_(none()))
        assert_that(Int().fromUnicode(''), is_(none()))

class TestDatetime(unittest.TestCase):

    def test_validate_wrong_type(self):
        field = ValidDatetime()
        assert_that(calling(field.validate).with_args(''),
                    raises(SchemaNotProvided))

class TestFieldValidationMixin(unittest.TestCase):

    def test_one_arg(self):
        field = FieldValidationMixin()
        field.__name__ = 'foo'

        ex = SchemaNotProvided('msg')
        try:
            field._reraise_validation_error(ex, 'value', _raise=True)
        except SchemaNotProvided:
            assert_that(ex.args, is_(('value', 'msg', 'foo')))

    def test_no_arg(self):
        field = FieldValidationMixin()
        field.__name__ = 'foo'

        ex = SchemaNotProvided()
        try:
            field._reraise_validation_error(ex, 'value', _raise=True)
        except SchemaNotProvided:
            assert_that(ex.args, is_(('value', '', 'foo')))

class TestRegex(unittest.TestCase):

    def test_regex(self):
        field = ValidRegularExpression('[bankai|shikai]', flags=0)
        assert_that(field.constraint("bankai"), is_(True))
        assert_that(field.constraint("shikai"), is_(True))
        assert_that(field.constraint("Shikai"), is_(False))
        assert_that(field.constraint("foo"), is_(False))
        field = ValidRegularExpression('[bankai|shikai]')
        assert_that(field.constraint("Shikai"), is_(True))
        assert_that(field.constraint("banKAI"), is_(True))



#!/usr/bin/env python

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

from hamcrest import is_
from hamcrest import is_not
from hamcrest import has_item
from hamcrest import not_none

from hamcrest import has_entry
from hamcrest import is_in
from hamcrest import assert_that
from hamcrest import has_property
does_not = is_not

import unittest

from zope import interface

from nti.schema.jsonschema import JsonSchemafier

from . import SchemaLayer


class TestConfiguredVocabulary(unittest.TestCase):

    layer = SchemaLayer

    def test_country_vocabulary(self):
        from zope.schema import Choice

        class IA(interface.Interface):
            choice = Choice(title="Choice",
                            vocabulary="Countries")

        o = object()

        choice = IA['choice'].bind(o)
        assert_that(choice.vocabulary, is_(not_none()))
        assert_that('us', is_in(choice.vocabulary))
        term = choice.vocabulary.getTermByToken('us')
        assert_that(term, has_property('value', "United States"))
        ext = term.toExternalObject()
        assert_that(ext, has_entry('flag', u'/++resource++country-flags/us.gif'))
        assert_that(ext, has_entry('title', 'United States'))

        schema = JsonSchemafier(IA).make_schema()
        assert_that(schema, has_entry('choice', has_entry('choices', has_item(ext))))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""


.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904

import unittest
from hamcrest import assert_that
from hamcrest import is_
from hamcrest import none
from hamcrest import has_property

from ..interfaces import InvalidValue

class TestInvalidValue(unittest.TestCase):

    def test_construct(self):
        v = InvalidValue()
        assert_that(v, has_property('value', none()))
        assert_that(v, has_property('field', none()))

        v = InvalidValue(value=1, field=1)
        assert_that(v, has_property('value', 1))
        assert_that(v, has_property('field', 1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""


.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904

import unittest
from hamcrest import assert_that
from hamcrest import is_
from hamcrest import has_key
from hamcrest import has_entry

from zope import interface
from .. import jsonschema
from ..field import DecodingValidTextLine

class TestJsonSchemafier(unittest.TestCase):

    def test_excludes(self):

        class IA(interface.Interface):

            def method():
                "A method"

            _thing = interface.Attribute("A private attribute")

            hidden = interface.Attribute("Hidden attribute")
            hidden.setTaggedValue(jsonschema.TAG_HIDDEN_IN_UI, True)

        schema = jsonschema.JsonSchemafier(IA).make_schema()
        assert_that(schema, is_({}))

    def test_readonly_override(self):
        class IA(interface.Interface):

            field = interface.Attribute("A field")

        schema = jsonschema.JsonSchemafier(IA, readonly_override=True).make_schema()
        assert_that(schema, has_entry('field', has_entry('readonly', True)))

    def test_ui_type(self):
        class IA(interface.Interface):

            field = interface.Attribute("A field")
            field.setTaggedValue(jsonschema.TAG_UI_TYPE, 'MyType')

        schema = jsonschema.JsonSchemafier(IA).make_schema()
        assert_that(schema, has_entry('field', has_entry('type', 'MyType')))

    def test_type_from_types(self):

        def _assert_type(t, name='field'):
            schema = jsonschema.JsonSchemafier(IA).make_schema()
            assert_that(schema, has_entry(name, has_entry('type', t)))
            return schema

        class IA(interface.Interface):

            field = interface.Attribute("A field")

            field2 = DecodingValidTextLine()

        IA['field']._type = str
        _assert_type('string')

        IA['field']._type = str,
        _assert_type('string')


        IA['field']._type = float
        _assert_type('float')

        IA['field']._type = float,
        _assert_type('float')

        IA['field']._type = int
        _assert_type('int')

        IA['field']._type = int,
        _assert_type('int')

        IA['field']._type = bool
        _assert_type('bool')

        schema = _assert_type('string', 'field2')
        assert_that(schema, has_entry('field2', has_entry('base_type', 'string')))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904


from nti.testing.layers import ZopeComponentLayer
from nti.testing.layers import ConfiguringLayerMixin

from zope.testing import cleanup

class SchemaLayer(ZopeComponentLayer,
                  ConfiguringLayerMixin):

    set_up_packages = ('nti.schema',)

    @classmethod
    def setUp(cls):
        cls.setUpPackages()

    @classmethod
    def tearDown(cls):
        cls.tearDownPackages()
        cleanup.cleanUp()

    @classmethod
    def testSetUp(cls, test=None):
        pass

    testTearDown = testSetUp

from zope.interface import Interface, classImplements

class IUnicode(Interface):
    "Unicode strings"

try:
    classImplements(unicode, IUnicode)
except NameError:
    classImplements(str, IUnicode)



#!/usr/bin/env python
from setuptools import setup, find_packages
import codecs

version = '1.1.0'

entry_points = {
}

def _read(fname):
    with codecs.open(fname, encoding='utf-8') as f:
        return f.read()

setup(
    name = 'nti.schema',
    version = version,
    author = 'Jason Madden',
    author_email = 'open-source@nextthought.com',
    description = ('Zope schema related support'),
    long_description = _read('README.rst'),
    license = 'Apache',
    keywords = 'zope schema',
    url = 'https://github.com/NextThought/nti.schema',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Framework :: Zope3',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'setuptools',
        'zope.schema',
        'zope.i18n',
        'zope.i18nmessageid',
        'zope.vocabularyregistry',
        'zope.deferredimport',
    ],
    extras_require={
        'test':[
            'nose2',
            'pyhamcrest',
            'nti.testing',
            'zope.testrunner',
        ],
        ':python_version == "2.7"': [
            # Not ported to Py3 yet; Plus, version 3 adds hard dep on
            # Products.CMFCore/Zope2 that we don't want.
            'plone.i18n < 3.0',
            'zope.browserresource', # Used by plone.i18n implicitly
        ]
    },
    namespace_packages=['nti'],
    entry_points=entry_points,
    test_suite='nose2.compat.unittest.collector'
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Computed attributes based on schema fields.

.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

import sys

from zope.schema import interfaces as sch_interfaces

# Re-export some things as part of our public API so we can
# later re-implement them locally if needed

from zope.schema.fieldproperty import FieldProperty
from zope.schema.fieldproperty import createFieldProperties
from zope.schema.fieldproperty import FieldPropertyStoredThroughField

from Acquisition import aq_base
from Acquisition.interfaces import IAcquirer

class AcquisitionFieldProperty(FieldProperty):
    """
    A field property that supports acquisition. Returned objects
    will be __of__ the instance, and set objects will always be the unwrapped
    base.
    """

    def __get__(self, instance, klass):
        result = super(AcquisitionFieldProperty, self).__get__(instance, klass)
        if instance is not None and IAcquirer.providedBy(result):  # even defaults get wrapped
            result = result.__of__(instance)
        return result

    def __set__(self, instance, value):
        super(AcquisitionFieldProperty, self).__set__(instance, aq_base(value))

class UnicodeConvertingFieldProperty(FieldProperty):
    """
    Accepts bytes input for the unicode property if it can be
    decoded as UTF-8. This is primarily to support legacy test cases
    and should be removed when all constants are unicode.
    """

    def __set__(self, inst, value):
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        super(UnicodeConvertingFieldProperty, self).__set__(inst, value)

def _find_schema_from_field(field):
    if not sch_interfaces.IObject.providedBy(field) and not hasattr(field, 'schema'):
        raise sch_interfaces.WrongType("Don't know how to get schema from %s" % field)
    return field.schema

def _make_adapter_set(klass):

    def __set__(self, inst, value):
        try:
            super(klass, self).__set__(inst, value)
        except sch_interfaces.SchemaNotProvided:
            try:
                value = self.schema(value)
            except TypeError:
                # Let's raise the better error when we call again
                pass

            super(klass, self).__set__(inst, value)

    return __set__

class AdaptingFieldProperty(FieldProperty):
    """
    Primarily for legacy support and testing, adds adaptation to an interface
    when setting a field. This is most useful when the values are simple literals
    like strings.
    """

    def __init__(self, field, name=None):
        self.schema = _find_schema_from_field(field)
        super(AdaptingFieldProperty, self).__init__(field, name=name)

AdaptingFieldProperty.__set__ =  _make_adapter_set(AdaptingFieldProperty)

class AdaptingFieldPropertyStoredThroughField(FieldPropertyStoredThroughField):
    """
    Primarily for legacy support and testing, adds adaptation to an interface
    when setting a field. This is most useful when the values are simple literals
    like strings.
    """

    def __init__(self, field, name=None):
        self.schema = _find_schema_from_field(field)
        super(AdaptingFieldPropertyStoredThroughField, self).__init__(field, name=name)

AdaptingFieldPropertyStoredThroughField.__set__ =  _make_adapter_set(AdaptingFieldPropertyStoredThroughField)

def createDirectFieldProperties(__schema, omit=(), adapting=False):
    """
    Like :func:`zope.schema.fieldproperty.createFieldProperties`, except
    only creates properties for fields directly contained within the
    given schema; inherited fields from parent interfaces are assummed
    to be implemented in a base class of the current class::

      class IA(interface.Interface):
          a = TextLine(title="a")

      class IB(IA):
          b = Object(IBaz)

      class A(object):
          createFieldProperties(IA)

      class B(object):
          createDirectFieldProperties(IB, adapting=True)


    :keyword adapting: If set to ``True`` (not the default), fields
        that implement :class:`.IObject` will use an :class:`AdaptingFieldProperty`.
    """

    __my_names = set(__schema.names())
    __all_names = set(__schema.names(all=True))

    __not_my_names = __all_names - __my_names
    __not_my_names.update(omit)

    # The existing implementation relies on getframe(1) to find the caller,
    # which is us. So we do the same and copy to the real caller
    __frame = None
    __before = None
    __before = list(locals().keys())
    createFieldProperties(__schema, omit=__not_my_names)

    __frame = sys._getframe(1)
    for k, v in locals().items():
        if k not in __before:
            if adapting and sch_interfaces.IObject.providedBy(__schema[k]):
                v = AdaptingFieldProperty(__schema[k])
            __frame.f_locals[k] = v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Vocabularies and factories for use in schema fields.

When this package is configured (via ``configure.zcml``) there will be a schema
vocabulary named ``Countries`` available::

  from nti.schema.field import Choice
  from zope import interface

  class IA(interface.Interface):
    choice = Choice(title="Choice",
                     vocabulary="Countries")



.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import component
from zope import interface

from zope.schema.vocabulary import SimpleTerm as _SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary as _SimpleVocabulary

try:
    from plone.i18n.locales.interfaces import ICountryAvailability as _ICountryAvailability
except ImportError:
    # Not on Py3
    class _ICountryAvailability(interface.Interface):
        def getCountries():
            ""
@interface.implementer(_ICountryAvailability)
class _CountryAvailabilityFallback(object):

    def getCountries(self):
        return {
            u'us' : {u'name' : 'United States', u'flag' : u'/++resource++country-flags/us.gif'},
        }



class CountryTerm(_SimpleTerm):
    """
    A titled, tokenized term representing a country. The
    token is the ISO3166 country code. The ``flag`` value is a
    browserresource path to an icon representing the country.
    """

    def __init__(self, *args, **kwargs):
        self.flag = kwargs.pop('flag', None)
        super(CountryTerm, self).__init__(*args, **kwargs)

    @classmethod
    def fromItem(cls, item):
        token, cdata = item
        value = cdata['name']
        title = value
        flag = cdata['flag']

        return cls(value, token, title, flag=flag)

    def toExternalObject(self):
        return { 'token': self.token,
                 'title': self.title,
                 'value': self.value,
                 'flag': self.flag }

class _CountryVocabulary(_SimpleVocabulary):
    """
    __contains__ is based on the token, not the value.
    """

    def __contains__(self, token):
        return token in self.by_token

def CountryVocabularyFactory(context):
    """
    A vocabulary factory, if plone.i18n is available.
    """
    countries = component.queryUtility(_ICountryAvailability) or _CountryAvailabilityFallback()
    return _CountryVocabulary([CountryTerm.fromItem(item) for item in countries.getCountries().items()])
